
// @module : Dummy_node
// @author : Michel Kinsy
 /*
 *  Using Heracles Multicore System under the copytright below:
 *  Copyright (c) 2011 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 *
 */
 
module Dummy_node #(parameter ID_BITS = 4, DATA_WIDTH = 32, IN_PORTS = 1, OUT_PORTS = 1, 
							VC_BITS = 1, VC_DEPTH_BITS = 4, EXTRA  = 2, TYPE_BITS = 2, 
							SWITCH_TO_SWITCH  = 1, RT_ALG = 0, ROW = 3, COLUMN = 3,

							TRAFFIC = 1, RATE = 5, FLITS_PER_PACKET = 8, STATS = 1024) (
							clock, reset, start, stop, 
							dest_ID,

							from_peripheral,
							from_peripheral_data, 
							from_peripheral_valid, 
							to_peripheral,
							to_peripheral_data, 
							to_peripheral_valid,

							ON, PROG, WEn, REn,
							router_ID, rt_flowID, rt_entry, 

							s_flits_in, s_valid_in, 
							s_flits_out, s_valid_out, 

							s_empty_in, s_full_in, 
							s_empty_out, s_full_out
);

localparam  VC_PER_PORTS  = (1 << VC_BITS); 
localparam  FLOW_BITS = (2*ID_BITS) + EXTRA;  
localparam  FLIT_WIDTH = FLOW_BITS + TYPE_BITS + VC_BITS + DATA_WIDTH;

localparam  PORTS = OUT_PORTS + 4*(SWITCH_TO_SWITCH);
localparam  PORTS_BITS = (PORTS <= 2) ? 1 : 
  						 (PORTS <= 4) ? 2 : (PORTS <= 8) ? 3 : 
  						 (PORTS <= 16) ? 4 : (PORTS <= 32) ? 5 : PORTS; 
localparam  RT_WIDTH = PORTS_BITS + VC_BITS + 1; 

input clock, reset, start, stop;
input ON, PROG, WEn, REn;

input  [1:0]   from_peripheral;
input  [31:0]  from_peripheral_data; 
input          from_peripheral_valid;
output [1:0]   to_peripheral;
output [31:0]  to_peripheral_data; 
output         to_peripheral_valid;

input [(FLIT_WIDTH * (SWITCH_TO_SWITCH *4)) - 1: 0] s_flits_in;
input [(SWITCH_TO_SWITCH *4) - 1: 0]s_valid_in; 

output [(FLIT_WIDTH * (SWITCH_TO_SWITCH *4)) - 1: 0] s_flits_out; 
output [(SWITCH_TO_SWITCH *4) - 1: 0] s_valid_out;

input [(((1 << VC_BITS) * (SWITCH_TO_SWITCH *4)) - 1): 0] s_empty_in; 
input [(((1 << VC_BITS) * (SWITCH_TO_SWITCH *4)) - 1): 0] s_full_in; 

output [(((1 << VC_BITS) * (SWITCH_TO_SWITCH *4)) - 1): 0] s_empty_out;
output [(((1 << VC_BITS) * (SWITCH_TO_SWITCH *4)) - 1): 0] s_full_out;  

input [(ID_BITS - 1): 0]    dest_ID;
input [(ID_BITS - 1): 0]    router_ID;
input [(FLOW_BITS - 1): 0] rt_flowID; 
input [RT_WIDTH - 1: 0] rt_entry;

wire [(ID_BITS - 1): 0]   core_ID =  router_ID;  
wire [(FLIT_WIDTH * IN_PORTS) -1:0] from_core_flit;
wire [IN_PORTS-1: 0] v_from_core;
wire [(VC_PER_PORTS * IN_PORTS) -1:0] from_core_empty;
wire [(VC_PER_PORTS * IN_PORTS) -1:0] from_core_full; 

wire  [(FLIT_WIDTH * OUT_PORTS) -1:0] to_core_flit;
wire  [OUT_PORTS-1: 0] v_to_core;
wire  [(VC_PER_PORTS * IN_PORTS) -1:0] to_core_empty;
wire  [(VC_PER_PORTS * IN_PORTS) -1:0] to_core_full;
// performance reporting
wire report; 

dummy_core #(TRAFFIC, RATE, DATA_WIDTH, IN_PORTS, OUT_PORTS, VC_PER_PORTS,
			ID_BITS, EXTRA, TYPE_BITS, ROW, COLUMN, FLITS_PER_PACKET, STATS)
                    injector (
						clock, reset, start, stop, 
						core_ID, dest_ID, 

						from_peripheral,
						from_peripheral_data, 
						from_peripheral_valid, 
						to_peripheral,
						to_peripheral_data, 
						to_peripheral_valid,

						from_core_flit, v_from_core, 
						from_core_empty, from_core_full,

						to_core_flit, v_to_core,
						to_core_empty, to_core_full, 
						report
);

router_wrapper #(RT_ALG, RT_WIDTH, FLOW_BITS, FLIT_WIDTH, IN_PORTS, 
						  OUT_PORTS, SWITCH_TO_SWITCH, VC_BITS, 
						  VC_DEPTH_BITS, ID_BITS, ROW, COLUMN) network_Interface (
						ON, clock, reset, PROG, WEn, REn,

						from_core_flit, v_from_core, 
						s_flits_in, s_valid_in, 

						to_core_flit, v_to_core,  
						s_flits_out, s_valid_out, 

						from_core_empty, from_core_full, s_empty_in, s_full_in, 
						to_core_empty, to_core_full, s_empty_out, s_full_out,  

						router_ID, rt_flowID, rt_entry, 
						report
);

endmodule
